
/*
 cc -g -c ../as/asinput.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

/* in case the OS needs distinction of files and sockets . Linux does not. */
#define closesocket close

#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/stcp.h"
#include "../s_tools/sconn.h"

#include "../as/asprog.h"
#include "../as/asqueue.h"
#include "../as/asinput.h"

#ifdef Asinput_use_curseS
#include <curses.h>
#else
#include <termios.h>
#endif /* Asinput_use_curseS */



/* ------------------------------- AsinconN --------------------------- */


int Asinconn_new(conn,input,type,sconn,prev,next,flag)
struct AsinconN **conn;
struct AsinpuT *input;
struct SconN *sconn;
int type;
struct AsinconN *prev,*next;
int flag;
{
 struct AsinconN *c;

 *conn= c= TSOB_FELD(struct AsinconN,1);
 if(c==NULL)
   return(-1);
 c->input= input;
 c->agent= input->agent;
 c->type= type|(1<<8)|(1<<9);
 c->bundle_max_size= Asinconn_default_bundle_sizE;
 c->bundle_overflow= 0;
 c->bundle_endmark= NULL;
 c->bundle_buffer= NULL;
 Sconn_refer(sconn,&(c->sconn),0);
 c->keep_pipe_alive_fd= -1;
 c->prev= prev;
 c->next= next;

#ifdef ImvdebuG 
 if(Asinconn_get_fd(c,0)>=0) {
   char line[1024];
   sprintf(line,"\ninput connection created.  fd=%ld , type=%-8.8X\n",
                 Asinconn_get_fd(c,0),c->type);
   Asagent_stdout(c->agent,line,1);
 }
#endif /* ImvdebuG */

 return(1);
failed:;
 Asinconn_destroy(conn,0);
 return(-1);
}


int Asinconn_destroy(conn,flag)
struct AsinconN **conn;
int flag;
/*
 bit0= also close underlying fd or destroy SconN 
*/
{
 struct AsinconN *c;
 int mode;

 c= *conn;
 mode= c->type&15;
 if(c==NULL)
   return(0);
 Sregex_string(&(c->bundle_endmark),NULL,0);
 Sregex_string(&(c->bundle_buffer),NULL,0);
 if(flag&1) {
   Sconn_destroy(&(c->sconn),0);
 }
 if(c->keep_pipe_alive_fd!=-1)
   close(c->keep_pipe_alive_fd);
 c->keep_pipe_alive_fd= -1;
 Asinconn_unlink(c,0);

#ifdef ImvdebuG 
 { char line[1024];
   sprintf(line,"\ninput connection destroyed.  fd=%ld , type=%-8.8X\n",
                 Asinconn_get_fd(c,1),c->type);
   Asagent_stdout(c->agent,line,1);
 }
#endif /* ImvdebuG */

 Smem_freE((char *) c);  /* *conn may have been altered by Asinconn_unlink() */ 
 *conn= NULL;
}


int Asinconn_unlink(conn,flag)
struct AsinconN *conn;
int flag;
{
 if(conn->input->conn_pt==conn)
   conn->input->conn_pt= NULL;
 if(conn->prev!=NULL)
   conn->prev->next= conn->next;
 if(conn->next!=NULL)
   conn->next->prev= conn->prev;
 conn->next= NULL;
 conn->prev= NULL;
 return(1);
}


int Asinconn_link(conn,prev,next,flag)
struct AsinconN *conn;
struct AsinconN *prev,*next;
int flag;
{

 Asinconn_unlink(conn,0);
 conn->prev= prev;
 if(prev!=NULL)
   prev->next= conn;
 conn->next= next;
 if(next!=NULL)
   next->prev= conn;
 return(1);
}


int Asinconn_set_bundle_endmark(conn,text,flag)
struct AsinconN *conn;
char *text;
int flag;
{
 if(Sregex_string(&(conn->bundle_endmark),text,0)<=0)
   return(-1);
 return(1);
}


int Asinconn_get_fd(conn,flag)
struct AsinconN *conn;
int flag;
/*
 bit1= get fd even if not valid
*/
{
 int fd;

 fd= Sconn_get_fd(conn->sconn,flag&2);
 return(fd);
}


int Asinconn_is_valid(conn,flag)
struct AsinconN *conn;
int flag;
{
 int ret;

 if(conn==NULL)
   return(0);
 ret= Sconn_is_valid(conn->sconn,0);
 return(ret);
}


int Asinconn_new_from_fd(conn,input,type,filename,fd,prev,next,flag)
struct AsinconN **conn;
struct AsinpuT *input;
int type;
char *filename;
int fd;
struct AsinconN *prev,*next;
int flag;
{
 int ret,stype;
 struct SconN *sconn;
 struct SaccesS *access;

 Asagent_get_access(input->agent,&access,0); 
 ret= Sconn_new(&sconn,filename,access,0);
 if(ret<=0)
   return(ret);
 stype= 3;
 if(type&(1<<10))
   stype= 2;
 else if(type&(1<<11))
   stype= 4;
 else if(type&(1<<12)) {
   if(type&(1<<13))
     stype= 1;
   else
     stype= 0;
 }
 ret= Sconn_attach_fd(sconn,stype,filename,fd,0);
 if(ret<=0)
   goto ex;
 ret= Asinconn_new(conn,input,type,sconn,prev,next,0);
 if(ret<=0)
   goto ex;
 return(1);
ex:
 Sconn_destroy(&sconn,0);
 return(ret);
}


int Asinconn_readchar(conn,reply,flag)
struct AsinconN *conn;
char *reply;
int flag;
/*
flag:
 bit0= no need to check wether fd is ready

return:
 <0 = error
  0 = eof
  1 = character read
  2 = no character available (would block)
*/
{
 int ret,count;

 ret= Sconn_readchar(conn->sconn,reply,flag);
 return(ret);
}


int Asinconn_readline(conn,line,flag)
struct AsinconN *conn;
char **line;
int flag;
/*
 return: <=0 error or EOF with empty line
         1   line complete (possibly completed by EOF)
         2   line not completed yet
*/
{
 int ret,l,bundle_open= 0;
 struct SaccesS *access;

 ret= Sconn_readline(conn->sconn,line,0);
 if(ret!=1 || (*line)==NULL)
   goto ex;
 Asagent_set_event_connection(conn->agent,Sconn_get_fd(conn->sconn,0),0);

 /* handle command bundling */
 if(conn->bundle_endmark!=NULL) if(conn->bundle_endmark[0]!=0) 
   bundle_open= 1;
 if(strcmp(*line,"-bundle:start")==0) {
   ret= Asagent_get_permission(conn->agent,"client","bundle",0);
   if(ret<0) 
     goto ex;
   if(ret==0)
     {ret= 2; goto ex;}
   if(Sregex_string(&(conn->bundle_endmark),"-bundle:end",0)<=0)
     {ret= -1; goto ex;}
   if(bundle_open)
     goto return_bundle_buffer;
   Sregex_string(line,NULL,0);
   ret= 2;
 } else if(!bundle_open) {
   ret= 1; goto ex;
 } else if(strcmp(*line,conn->bundle_endmark)==0) {
   Sregex_string(&(conn->bundle_endmark),NULL,0);
return_bundle_buffer:;
   if(conn->bundle_overflow) {
     if(Sregex_string(line,"",0)<=0)
       {ret= -1; goto ex;};
     Asagent_stderr(conn->agent,
                      "Buffer overflow with -bundle. Bundle was discarded.",0);
     ret= 0;
   } else {
     if(Sregex_string(line,"-bundle:start\n",0)<=0)
       {ret= -1; goto ex;}
     if(conn->bundle_buffer!=NULL) {
       if(Sregex_string(line,conn->bundle_buffer,1)<=0)
         {ret= -1; goto ex;}
       if(Sregex_string(line,"\n",1)<=0)
         {ret= -1; goto ex;}
     }
     if(Sregex_string(line,"-bundle:end",1)<=0)
       {ret= -1; goto ex;}
     ret= 1;
   }
   Sregex_string(&(conn->bundle_buffer),NULL,0);
   conn->bundle_overflow= 0;
 } else {
   if(!conn->bundle_overflow) {
     l= 0;
     if(conn->bundle_buffer!=NULL)
       l= strlen(conn->bundle_buffer);
     else
       conn->bundle_overflow= 0;
     if(l+strlen(*line)+1>conn->bundle_max_size) {
       Asagent_stderr(conn->agent,
                  "Buffer overflow with -bundle. Bundle will be discarded.",1);
       Sregex_string(&(conn->bundle_buffer),NULL,0);
       conn->bundle_overflow= 1;
     } else {
       if(l>0)
         if(Sregex_string(&(conn->bundle_buffer),"\n",1)<=0)
           {ret= -1; goto ex;}
       if(Sregex_string(&(conn->bundle_buffer),*line,1)<=0)
         {ret= -1; goto ex;}
     }
   }
   ret= 2;
   Sregex_string(line,NULL,0);
 }
ex:;
 Asagent_set_event_connection(conn->agent,0,1);
 return(ret);
}


int Asinconn_determine_user(conn,username,flag)
struct AsinconN *conn;
char **username;
int flag;
{
 int ret;

 if(conn->sconn==NULL)
   goto external_anonymous;
 ret= Sconn_get_username(conn->sconn,username,0);
 if(ret<0)
   goto ex;
 if(ret==0 || *username==NULL) {
external_anonymous:;
   if(Sregex_string(username,"external_anonymous",0)<=0)
     {ret= -1;goto ex;}
 }
 ret= 1;
ex:; 
 return(ret);
}


/* ------------------------------- AsinpuT --------------------------- */


int Asinput_new(input,queue,type,progname,flag)
struct AsinpuT **input;
struct AsqueuE *queue;
int type;
char *progname;
int flag;
/*
 bit0= do not start input
*/
{
 int ret;
 struct AsinpuT *i;
 struct SconN *conn;

 *input= i= TSOB_FELD(struct AsinpuT,1);
 if(i==NULL)
   return(-1);
 i->queue= queue;
 i->agent= queue->agent;
 i->type= type;
 i->initialized= 0;
 i->termios_orig= NULL;
 i->char_serv= NULL;
 i->line_serv= NULL;
 i->connections= NULL;
 i->conn_pt= NULL;
 i->prog= NULL;
 i->reply= NULL;
 i->reply_count= 0;
 i->reply_type= 0;
 i->from_connection= -1;
 i->username= NULL;

 ret= Asinconn_new_from_fd(&(i->connections),i,(1<<11),"dummy",
                            -1,NULL,NULL,0);
 if(ret<=0)
   goto failed;

 ret= Asprog_new(&(i->prog),i,50,0);
 if(ret<=0)
   goto failed;
 if(progname[0]==0){
   /* ret= Asprog_default_kbd(i->prog,0); */;
 }else{
   ret= Asprog_read_kbd_v(i->prog,progname,0);
 }
 if(ret<=0)
   goto failed;
 if(!(flag&1)) {
   ret= Asinput_start(i,0);
   if(ret<=0)
     goto failed;
 }
 return(1);
failed:;
 Asinput_destroy(input,0);
 return(ret);
}


int Asinput_destroy(input,flag)
struct AsinpuT **input;
int flag;
{
 if(*input==NULL)
   return(0);

#ifndef Asinput_use_curseS
 if((*input)->termios_orig!=NULL) {
   tcsetattr(0,TCSANOW,(*input)->termios_orig);
   Smem_freE((char *) (*input)->termios_orig);
 }
#endif

 Asinput_destroy_connection(*input,-2,1);
 Asprog_destroy(&((*input)->prog),0);
 Asinput_destroy_reply(*input,0);
 Sregex_string(&((*input)->username),NULL,0);
 Asinput_stop(*input,0);
 Smem_freE((char *) *input);
 *input= NULL;
 return(1);
}


int Asinput_destroy_reply(input,flag)
struct AsinpuT *input;
int flag;
{
 int i;

 if(input->reply==NULL){
   input->reply_count= 0;
   return(0);
 }
 for(i=0;i<input->reply_count;i++)
   if(input->reply[i]!=NULL)
     Smem_freE((char *) input->reply[i]);
 Smem_freE((char *) input->reply);
 input->reply= NULL;
 input->reply_count= 0;
 return(1);
}


int Asinput_set_internal(input,value,flag)
struct AsinpuT *input;
int value;
int flag;
{
 input->type= (input->type&~1)|!!value;
 return(1);
}


int Asinput_set_reply(input,raw_reply,flag)
struct AsinpuT *input;
char *raw_reply;
int flag;
/*
 bit0= non ESC : use raw_reply without interpretation
 bit1= ESC : use raw_reply without interpretation
*/
{
 char *cpt,*reply,*buff= NULL,*agent_reply=NULL,**replies;
 int ret,l= 0,i;

 input->reply_type= 0;
 Asinput_destroy_reply(input,0);
 if(raw_reply[0]==27 && raw_reply[1]==0) {
   if(flag&2) {
     ret= Asinput_set_replyline(input,raw_reply,0);
     if(ret<=0)
       goto ex;
   } else {
     do{
       ret= Asinput_fetch_token(input,1|4|32|128);
       if(ret<0)
         goto ex;
     } while(ret>0);
     input->reply_type= 1; 
     ret= Asinput_set_replyline(input,"CANCEL",0);
     if(ret<=0)
       goto ex;
   }
 } else if(flag&1) {
   ret= Asinput_set_replyline(input,raw_reply,0);
   if(ret<=0)
     goto ex;
 } else if(raw_reply[0]=='@') {
   ret= Asagent_fetch_line(input->agent,"enter command line :",
                            &agent_reply,0);
   if(ret<=0)
     goto ex;
   if(agent_reply[0]!='-' && agent_reply[0]!=':' && agent_reply[0]!=0) {
     if(Sregex_string(&buff,"-",0)<=0)
       {ret= -1; goto ex;}
   }
   if(Sregex_string(&buff,agent_reply,1)<=0)
     {ret= -1; goto ex;}
   ret= Asinput_set_replyline(input,buff,0);
   if(ret<=0)
     goto ex;
 } else {
   ret= Asprog_interpreter(input->prog,raw_reply,0);
   if(ret<=0)
     goto ex;
   replies= TSOB_FELD(char *,input->prog->reply->reply_count+1);   
   if(replies==NULL)
     {ret= -1; goto ex;}
   for(i=0;i<input->prog->reply->reply_count+1;i++)
     replies[i]= NULL;
   for(i=0;i<input->prog->reply->reply_count;i++){
     if(Sregex_string(&(replies[i]),input->prog->reply->reply[i],0)<=0)
       {ret= -1; goto ex;}
   }
   input->reply_count= input->prog->reply->reply_count;
   input->reply= replies;
 }
 ret= 1;
ex:;
 Sregex_string(&agent_reply,NULL,0);
 Sregex_string(&buff,NULL,0);
 return(ret);
}


int Asinput_default_kbd(input,flag)
struct AsinpuT *input;
int flag;
{
 int ret;

 ret= Asprog_default_kbd(input->prog,flag);
 return(ret);
}


int Asinput_set_replyline(input,line,flag)
struct AsinpuT *input;
char *line;
int flag;
{
 int ret,count= 1,i;
 char **replies,*cpt,*line_token= NULL;
 
 /* count newlines */
 cpt= line;
 for(count= 0;cpt!= NULL;count++)
   if((cpt= strchr(cpt,'\n'))!=NULL)
     cpt++;

 Asinput_destroy_reply(input,0);
 replies= TSOB_FELD(char *,count);   
 if(replies==NULL)
   {ret= -1; goto ex;}
 for(i=0;i<count;i++)
   replies[i]= NULL;
 /* copy replylines */
 cpt= line;
 for(i=0;i<count;i++) {
   ret= Sregex_cut_neck(cpt,'\n',&line_token,&cpt,0);
   if(ret<=0)
     goto ex;
   ret= Sregex_string(&(replies[i]),line_token,0);
   if(ret<=0)
     goto ex;
 }
 input->reply_count= count;
 input->reply= replies;
 ret= 1;
ex:;
 Sregex_string(&line_token,NULL,0);
 return(ret);
}


int Asinput_get_replyline(input,line,flag)
struct AsinpuT *input;
char **line;
int flag;
{
 int i,ret;

 if(Sregex_string(line,"",0)<=0)
   return(-1);
 for(i=0;i<input->reply_count;i++){
   if(input->reply[i]==NULL)
 continue;
   if(strlen(*line)>0)
     if(Sregex_string(line," ",1)<=0)
       return(-1);
   if(Sregex_string(line,input->reply[i],1)<=0)
     return(-1);
 }
 return(1);
}


int Asinput_count_services(input,flag)
struct AsinpuT *input;
int flag;
/*
 bit0= count internal input as a service
*/
{
 struct AsinconN *c;
 int count= 0,ret;

 if(flag&1) 
   if(input->type&1)
     count++;
 c= input->connections;
 for(c= input->connections;c!=NULL;c= c->next)
   if(!(c->type&2048))
     if((c->type&15)!=3 && (c->type&15)!=4 && (c->type&15)!=5)
       count++;
 return(count);
}


int Asinput_add_connection(input,conn,flag)
struct AsinpuT *input;
struct AsinconN *conn;
int flag;
{
 struct AsinconN *c;

 c= input->connections;
 while(c!=NULL){
   if(c->next==NULL)
 break;
   c= c->next;
 }
 if(c==NULL) {
   input->connections= conn;
   Asinconn_link(conn,NULL,NULL,0);
 } else {
   Asinconn_link(conn,c,NULL,0);
 }
 return(1);
}


int Asinput_new_inconn(input,type,filename,fd,keep_pipe_alive_fd,flag)
struct AsinpuT *input;
int type;
char *filename;
int fd;
int keep_pipe_alive_fd;
int flag;
{
 int ret;
 struct AsinconN *conn= NULL;

 ret= Asinconn_new_from_fd(&conn,input,type,filename,fd,NULL,NULL,0);
 if(ret<=0)
   return(ret);
 conn->keep_pipe_alive_fd= keep_pipe_alive_fd;
 ret= Asinput_add_connection(input,conn,0);
 return(ret);
}


int Asinput_sconn_to_inconn(input,type,sconn,flag)
struct AsinpuT *input;
int type;
struct SconN *sconn;
int flag;
{
 int ret;
 struct AsinconN *conn= NULL;

 ret= Asinconn_new(&conn,input,type,sconn,NULL,NULL,0);
 if(ret<=0)
   return(ret);
 ret= Asinput_add_connection(input,conn,0);
 return(ret);
}


int Asinput_destroy_connection(input,fd,flag)
struct AsinpuT *input;
int fd;                   /* -2 = all */             
int flag;
/*
 bit0= (together with fd= -2) also destroy first (dummy) connection of input
*/
{
 struct AsinconN *c,*next,*prev;

 c= input->connections;
 while(c!=NULL){
   next= c->next;
   if(next==NULL)
 break;
   c= next;
 }
 while(c!=NULL && c!=input->connections){
   prev= c->prev;
   if(fd==-2 || Asinconn_get_fd(c,0)==fd) {
     Asagent_notify_closedconn(input->agent,Asinconn_get_fd(c,2),1);
     Asinconn_destroy(&c,1);
   }
   c= prev;
 }
 if((flag&1)&&fd==-2)
   Asinconn_destroy(&(input->connections),1);
 return(1);
}


int Asinput_find_inconn(input,conn,id,fd,flag)
struct AsinpuT *input;
struct AsinconN **conn;
char *id;
int fd;
int flag;
/*
 bit0= 'id' is a cfd filename (else: a sconn-id)
 bit1= with bit 0 : filename of outbound cfd (else inbound)
*/
{
 struct AsinconN *c,*next,*prev;
 char *conn_id= NULL;
 int ret;

 *conn= NULL;
 c= input->connections;
 while(c!=NULL){
   next= c->next;
   if(id!=NULL){
     if(flag&1) {
       ret= Sconn_get_filename(c->sconn,&conn_id,(flag>>1)&1);
     } else {
       ret= Sconn_get_id(c->sconn,&conn_id,0);
     }
     if(ret<=0)
       goto ex;
     if(strcmp(id,conn_id)!=0)
       goto next_c;
   }
   if(fd>=0) {
     if(Sconn_get_fd(c->sconn,0)!=fd && Sconn_get_fd(c->sconn,1)!=fd)
       goto next_c;
   }
   *conn= c;
   ret= 1;
   goto ex;
next_c:;
   if(next==NULL)
 break;
   c= next;
 }
 ret= 0;
ex:
 Sregex_string(&conn_id,NULL,0);
 return(ret);
}


int Asinput_find_sconn(input,sconn,id,fd,flag)
struct AsinpuT *input;
struct SconN **sconn;
char *id;
int fd;
int flag;
/*
 bit0= 'id' is a cfd filename
 bit1= with bit 0 : filename of outbound cfd (else inbound)
*/
{
 int ret;
 struct AsinconN *c= NULL;

 *sconn= NULL;
 ret= Asinput_find_inconn(input,&c,id,fd,flag&3);
 if(ret<=0)
   return(ret);
 *sconn= c->sconn;
 return((*sconn)!=NULL);
}


int Asinput_set_bundle_endmark(input,fd,text,flag)
struct AsinpuT *input;
int fd;
char *text;
int flag;
{
 int ret;
 struct AsinconN *c= NULL;

 if(fd<0) {
   Asagent_stderr(input->agent,
                   "Bundling not possible with internal input channel",1);
   return(0);
 }
 ret= Asinput_find_inconn(input,&c,NULL,fd,0);
 if(ret<=0)
   return(ret);
 ret= Asinconn_set_bundle_endmark(c,text,0);
 return(ret);
}


int Asinput_start(input,flag)
struct AsinpuT *input;
int flag;
{
 if(input->initialized)
   Asinput_stop(input,flag);
 if((input->type&1)){
#ifdef Asinput_use_curseS
   initscr();
   cbreak();
   noecho();
   if(Asinput_curses_halfdelaY>0)
     halfdelay(Asinput_curses_halfdelaY);
   refresh();
   endwin();
#else
   ;
#endif /* ! Asinput_use_curseS */
   input->initialized|= 1;
 }
 return(1);
}


int Asinput_start_tcpserv(input,servername,portnumber,external_s_mask,
                           type,flag)
struct AsinpuT *input;
char *servername;
int portnumber;
int external_s_mask;
int type;
int flag;
{
 int ret;
 struct StcpserV *serv,**target;
 struct AsinconN *conn=NULL;
 
 Asinput_end_tcpserv(input,type,0);
 if(type==1) {
   target= &(input->char_serv);
 } else if(type==2) {
   target= &(input->line_serv);
 } else
   return(-1);
 ret= Stcpserv_new(target,servername,portnumber,external_s_mask,0);
 if(ret<=0)
   return(ret);
 type|= (1<<10); /* mark as master socket */

 ret= Asinconn_new_from_fd(&conn,input,type,NULL,(*target)->server_socket,
                        NULL,NULL,0);
 if(ret<=0)
   return(ret);
 ret= Asinput_add_connection(input,conn,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(ret<=0){
   Asinconn_destroy(&conn,1);
   Stcpserv_destroy(target,0);
 }
 return(ret);
}


int Asinput_end_tcpserv(input,type,flag)
struct AsinpuT *input;
int type;
int flag;
{
 int ret;
 struct StcpserV **target;
 struct AsinconN *conn=NULL;
 
 if(type==1) {
   target= &(input->char_serv);
 } else if(type==2) {
   target= &(input->line_serv);
 } else
   return(-1);
 if(*target==NULL)
   return(0);
 /* look in connection list for fd of this server */
 conn= input->connections;
 while(conn!=NULL) {
   if(conn->type&(1<<10))
     if((*target)->server_socket==Asinconn_get_fd(conn,0))
 break;
   conn= conn->next;
 }
 /* delete this connection */
 if(conn!=NULL){
   Asagent_notify_closedconn(input->agent,Asinconn_get_fd(conn,0),0);
   Asinconn_destroy(&conn,0);
 }
 Stcpserv_destroy(target,0);

 ret= 1;
ex:;
 return(ret);
}


int Asinput_stop(input,flag)
struct AsinpuT *input;
int flag;
{
 if(input->initialized&1) {
#ifdef Asinput_use_curseS
   endwin();
#endif /* Asinput_use_curseS */
 }
 if(input->char_serv!=NULL)
   Asinput_end_tcpserv(input,1,0);
 if(input->line_serv!=NULL)
   Asinput_end_tcpserv(input,2,0);

 /* >>> what about  input->connections ? */
 
 input->initialized= 0;
 return(1);
}


int Asinput_show_text(input,flag)
struct AsinpuT *input;
int flag;
{
 if(!(input->initialized&1))
   return(0);
#ifdef Asinput_use_curseS
 refresh();
 endwin();
#else
   ;
#endif /* ! Asinput_use_curseS */
 return(1);
}


int Asinput_text_ttysafe(input,text,flag)
struct AsinpuT *input;
char *text;
int flag;
{
 int i,l;

 l= strlen(text);
 for(i=0;i<l;i++) {
   if((text[i]&127)>=32)
 continue;
   if(strchr("\a\b\t\n\v\f\r",text[i])!=NULL)
 continue;
   text[i]= '_';
 }
 return(1);
}


int Asinput_fprintf(input,fp,buffer,buffer_size,in_text,flag)
struct AsinpuT *input;
FILE *fp;
char *buffer;
int buffer_size;
char *in_text;
int flag;
/*
 bit0= append \n
 bit2= flush buffer 
*/
{
 int ret;
 char *text= NULL;

 if(Sregex_string(&text,in_text,0)<=0)
   {ret= -1; goto ex;}
 Asinput_text_ttysafe(input,text,0);
 if(strlen(buffer)+strlen(text)>=buffer_size) {
   fprintf(fp,"%s",buffer);
   buffer[0]= 0;
   Asinput_show_text(input,0);
 }
 if(strlen(text)>=buffer_size) 
   fprintf(fp,"%s",text);
 else
   strcat(buffer,text);
 if(flag&5) {
   if(flag&1) 
     fprintf(fp,"%s\n",buffer);
   else
     if(strlen(buffer)>0)
       fprintf(fp,"%s",buffer);
   buffer[0]= 0;
   Asinput_show_text(input,0);
 }
 ret= 1;
ex:;
 Sregex_string(&text,NULL,0);
 return(ret);
}


int Asinput_stdout(input,text,flag)
struct AsinpuT *input;
char *text;
int flag;
/*
 bit0= append \n
 bit2= flush buffer 
*/
{
 static char buffer[SfileadrL]= {""};
 int buffer_size= SfileadrL;

 Asinput_fprintf(input,stdout,buffer,buffer_size,text,flag&5);
 return(1);
}


int Asinput_stderr(input,text,flag)
struct AsinpuT *input;
char *text;
int flag;
/*
 bit0= append \n
 bit2= flush buffer 
*/
{
 static char buffer[SfileadrL]= {""};
 int buffer_size= SfileadrL;

 Asinput_fprintf(input,stderr,buffer,buffer_size,text,flag&5);
 return(1);
}


int Asinput_choose_conn(input,microsec,flag)
struct AsinpuT *input;
unsigned int microsec;
int flag;
/*
 bit0= only select internal and line input that subscribed 'ask'
 bit1= only select mode 3 and 4 (stdout, stderr pipes)
 bit2= do not select mode 5 (client view input)
*/
{
 int ret= 0,ready,fd_max,mode,fd;
 struct AsinconN *conn,*start_pt;
 char *dummy= NULL; /* unused pointer to struct Imvview */
 struct timeval wt;
 fd_set rds,wts,exs,enabled;

 wt.tv_sec=  microsec/1000000;
 wt.tv_usec= microsec%1000000;
 FD_ZERO(&rds);
 FD_ZERO(&wts);
 FD_ZERO(&exs);
 FD_ZERO(&enabled);
    
 conn= input->connections;
 if(conn==NULL)
   return(0);
 start_pt= input->conn_pt;
 if(start_pt==NULL)
   start_pt= input->connections;
 start_pt= start_pt->next;
 if(start_pt==NULL)
   start_pt= input->connections;
 fd_max= -1;
 while(conn!=NULL) {
   fd= Asinconn_get_fd(conn,0);
   mode= conn->type&15;
   if(fd<0 || (conn->type&(1<<11))) 
     goto set_next;
   if((flag&2)&&mode!=3&&mode!=4)
     goto set_next;
   if((flag&4)&&mode==5)
     goto set_next;
   if(!(flag&1) || mode==0 
       || (mode==2 && Asagent_find_view(input->agent,&dummy,
                                         "","","",fd,"ask",1)>0)
       ) {
     FD_SET(fd,&rds);
     FD_SET(fd,&exs);
     FD_SET(fd,&enabled);
     if(fd>fd_max)
        fd_max= fd;
     if(Sconn_has_bytes(conn->sconn,0))
       {wt.tv_sec= wt.tv_usec= 0;}
   }
set_next:;
   conn= conn->next;
 }
 if(fd_max<0)
   return(0);
 ready= select(fd_max+1,&rds,&wts,&exs,&wt);
 if(ready<=0) {
   FD_ZERO(&rds);
   FD_ZERO(&wts);
   FD_ZERO(&exs);
 }
 conn= start_pt;
 do{
   fd= Asinconn_get_fd(conn,0);
   if(!(conn->type&(1<<11))) {
     if(FD_ISSET(fd,&rds) || FD_ISSET(fd,&exs)
        || (FD_ISSET(fd,&enabled) && Sconn_has_bytes(conn->sconn,0))) {
       input->conn_pt= conn;
       ret= 1;
       goto ex;
     }
   }
   conn= conn->next;
   if(conn==NULL)
     conn= input->connections;
 } while(conn!=start_pt);
 ret= 0;
ex:;
 return(ret);
}


int Asinput_accept_conn(input,type,flag)
struct AsinpuT *input;
int type;
int flag;
{
 struct AsinconN *conn=NULL;
 int fd= -1,ret;

 if(type==1){
   ret= Stcpserv_accept_fd(input->char_serv,&fd,0);
 } else if(type==2) {
   ret= Stcpserv_accept_fd(input->line_serv,&fd,0);
 } else {
   return(-1);
 }
 if(ret<=0)
   return(ret);
 ret= Asinconn_new_from_fd(&conn,input,type,NULL,fd,NULL,NULL,0);
 if(ret<=0)
   return(ret);
 ret= Asinput_add_connection(input,conn,0);
 if(ret<=0)
   return(ret);

 ret= 1;
ex:;
 if(ret<=0)
   Asinconn_destroy(&conn,1); 
 return(ret);
}


int Asinput_readchar_internal(input,gotchar,flag)
struct AsinpuT *input;
char *gotchar;
int flag;
/*
 bit0= set blocking mode according to bit1 (halfdelay else, if possible)
 bit1= blocking, non blocking else 
return:
 -1 = fatal error
  0 = no character read
  1 = character read
*/
{
 int got,ret;
 unsigned int microsec;

#ifndef Asinput_use_curseS
 static struct termios termios_p;

 if(input->termios_orig==NULL) {
   input->termios_orig= TSOB_FELD(struct termios,1);   
   if(input->termios_orig==NULL)
     return(-1);
   tcgetattr(0,input->termios_orig);
 }
#endif /* ! Asinput_use_curseS */

 if(!input->initialized)
   return(0);
 if(input->type&1){
#ifdef Asinput_use_curseS
   initscr();
   cbreak();
   noecho();
   if((flag&1) || (Asinput_curses_halfdelaY<=0)) {
     if(flag&2)
       nodelay(stdscr,FALSE);
     else
       nodelay(stdscr,TRUE);
   } else
     halfdelay(Asinput_curses_halfdelaY);
   errno= 0;
   got= getch();
   endwin();
   if(got==ERR) {
     if(errno!=0)
       return(-1);
     else 
       return(0);
   }
   gotchar[0]= got;

#else  /* Asinput_use_curseS */

   microsec= 100000;
   if((flag&1)&&!(flag&2))
     microsec= 0;
   ret= tcgetattr(0,&termios_p);
   termios_p.c_lflag&= ~(ICANON|ECHO);
   tcsetattr(0,TCSANOW,&termios_p);
try_again:;
   ret= Sfile_fd_is_ready(0,microsec,1);
   if(ret<0)
     goto ex;
   if(ret==0) {
     if((flag&1)&&(flag&2))
       goto try_again;
     ret= 0;
     goto ex;
   }
   ret= read(0,gotchar,1);
   if(ret<=0) {
     if(errno==EAGAIN)
       if((flag&1)&&(flag&2))
         goto try_again;
     ret= -1;
     goto ex;
   }
   if(input->termios_orig!=NULL)
     tcsetattr(0,TCSANOW,input->termios_orig);
#endif /* ! Asinput_use_curseS */
 } else {
   ret= 0;
   goto ex;
 }
 ret= 1;
ex:;
#ifndef Asinput_use_curseS
 if(input->termios_orig!=NULL)
   tcsetattr(0,TCSANOW,input->termios_orig);
#endif
 return(ret);
}


int Asinput_readline_internal(input,result_line,flag)
struct AsinpuT *input;
char **result_line;
int flag;
/*
 bit0= do not print question
 bit1= unconditionally blocking read
*/
{
 char line[SfileadrL];
 int ret,l;

 if(!(flag&2))
   if(!Sfile_fd_is_ready(fileno(stdin),0,1))
     return(0);
 ret= 0;
 if(fgets(line,sizeof(line),stdin)==NULL) {
   if(feof(stdin)) {
     Asagent_stderr(input->agent,"EOF at standard input",1);
   } else {
     sprintf(line,"Error %ld at standard input.\n%s",errno,strerror(errno));
     Asagent_stderr(input->agent,line,1);
   }
   return(-1);
 }
 l= strlen(line);
 if(l>0) if(line[l-1]=='\n') {
   line[--l]= 0;
   if(l>0) if(line[l-1]=='\r')
     line[--l]= 0;
 }
 if(Sregex_string(result_line,line,0)<=0)
   return(-1);
 return(1);
}


int Asinput_fetch_line(input,flag)
struct AsinpuT *input;
int flag;
/*
 bit0= do not print question
*/
{
 int ret,l;

#ifdef OLD_INPUT_FETCH_LINE
 char line[SfileadrL];

 if(fgets(line,sizeof(line),stdin)==NULL) {

   if(feof(stdin)) {
     Asagent_stderr(input->agent,"EOF at standard input",1);
   } else {
     sprintf(line,"Error %ld at standard input.\n%s",errno,strerror(errno));
     Asagent_stderr(input->agent,line,1);
   }
   return(-1);
 }
#else /* OLD_INPUT_FETCH_LINE */
 char *line= NULL;

 do {
   ret= Asinput_fetch_token(input,2|8);
   if(ret<0)
     goto ex;
 } while(ret==0);
 ret= Asinput_get_replyline(input,&line,0);
 if(ret<=0)
   goto ex;
 
#endif /* ! OLD_INPUT_FETCH_LINE */

 l= strlen(line);
 if(l>0)
   if(line[l-1]=='\n') {
     line[l-1]= 0;
     if(l>1)
       if(line[l-2]=='\r')
         line[l-2]= 0;
   }
 ret= Asinput_set_replyline(input,line,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
#ifndef OLD_INPUT_FETCH_LINE
 Sregex_string(&line,NULL,0);
#endif /* ! OLD_INPUT_FETCH_LINE */
 return(ret);
}


int Asinput_fetch_token(input,flag)
struct AsinpuT *input;
int flag;
/*
 bit0= set blocking mode according to bit1 (halfdelay else, if possible)
 bit1= blocking, non blocking else 
 bit2= do not convert reply character by prog
 bit3= this looks for requested line input
 bit4= only look for modes 3 and 4 (stdout,stderr pipes)
 bit5= do not look for mode 5
 bit6= only accept internal input
 bit7= throw away mode, do not interpret or set reply line (exept mode 5)
return:
 -1 = fatal error
  0 = no token read
  1 = token read
  2 = token read, request do deliver fast
*/
{
 int ret= 0,hflag= 0, mode,count,linemode=0,choosemode,setmode;
 unsigned int choice_wait= 0;
 struct AsinconN *conn= NULL;
 char buf[2],*input_line= NULL,line[81];

 linemode= !!(flag&8);
 choosemode= (flag>>3)&7;
 setmode= (flag>>3)&3;

try_again:;
 input->from_connection= -1;
 Sregex_string(&(input->username),NULL,0);
 if((input->type&1) && !(flag&16)) {
   if(linemode) {
     hflag= 0;
     ret= Asinput_readline_internal(input,&input_line,hflag);
   } else {
     if(Sregex_string(&input_line," ",0)<=0)     /* allocate 2 char memory */
       {ret= -1; goto ex;}
     hflag= (flag&3)!=3 && (flag&3)!=0;
     ret= Asinput_readchar_internal(input,input_line,hflag);
   }
   if(ret<0)
     goto ex;
   if(ret>0){
     if(flag&128)
       {ret= 1; goto ex;}
     ret= Asinput_set_reply(input,input_line,setmode||(flag&4));
     if(Sregex_string(&(input->username),"internal_owner",0)<=0)
       {ret= -1; goto ex;}
     goto ex;
   }
   choice_wait= 0;
 } else {
   if((flag&1)&&!(flag&2)) 
     choice_wait= 0;
   else
     choice_wait= 100000;
 }

choose_conn:;
 if(flag&64)
   goto no_request;
 input->from_connection= -1;
 Sregex_string(&(input->username),NULL,0);
 ret= Asinput_choose_conn(input,choice_wait,choosemode);
 if(ret<0)
   goto ex;
 if(ret==0)
   goto no_request;
 conn= input->conn_pt;
 input->from_connection= Asinconn_get_fd(conn,0);
 ret= Asinconn_determine_user(conn,&(input->username),0);
 if(ret<0)
   goto ex;
 mode= conn->type&15;
 if(mode==0) {
   if(linemode) {
     ret= Asinput_readline_internal(input,&input_line,0);
   } else {
     if(Sregex_string(&input_line," ",0)<=0)     /* allocate 2 char memory */
       {ret= -1; goto ex;}
     ret= Asinput_readchar_internal(input,input_line,1);
   }
   if(ret<0)
     goto ex;
   if(ret>0){
     if(flag&128)
       {ret= 1; goto ex;}
     ret= Asinput_set_reply(input,input_line,setmode||(flag&4));
     goto ex;
   }
 } else if(mode==1) {
   if(conn->type&1024) {
     Asinput_accept_conn(input,1,0);
   } else {
     count= 0;
     ret= Asinconn_readchar(conn,buf,1);
     if(ret<=0){
       Asagent_notify_closedconn(input->agent,Asinconn_get_fd(conn,2),0);
       Asinconn_destroy(&(input->conn_pt),1);
       goto choose_conn;
     } else if(ret==1)
       count= 1;
     if(count==0)
       goto choose_conn;
     buf[1]= 0;  
     if(flag&128)
       {ret= 1; goto ex;}
     ret= Asinput_set_reply(input,buf,!!(flag&4));
     if(ret!=0)
       goto ex;
   }
 } else if(mode==2) {
   if(conn->type&1024) {
     Asinput_accept_conn(input,2,0);
   } else {
     ret= Asinconn_readline(conn,&input_line,0);
     if(ret<=0 || !Asinconn_is_valid(conn,0)){
       Asagent_notify_closedconn(input->agent,Asinconn_get_fd(conn,2),0);
       Asinconn_destroy(&(input->conn_pt),1);
       if(ret<=0)
         goto choose_conn;
     }
     if(ret==2)
       goto choose_conn;
     if(flag&128)
       {ret= 1; goto ex;}
     ret= Asinput_set_reply(input,input_line,1);
     if(ret!=0)
       goto ex;
   }
 } else if(mode==3 || mode==4) {
   ret= Asinconn_readline(conn,&input_line,0);
   if(ret<=0||ret==2)
     goto choose_conn;
   if(input_line[0]) {
     if(mode==3)
       Asagent_stdout(input->agent,input_line,1);
     else
       Asagent_stderr(input->agent,input_line,1);
   }
   goto choose_conn;
 } else if(mode==5) {
   sprintf(line,"-input_ready_fd:%ld",Asinconn_get_fd(conn,0));
   ret= Asinput_set_reply(input,line,1);
   if(ret<=0)
     goto ex;
   ret= 2; goto ex;
 }
no_request:;
 if((flag&3)==3) /* blocking mode */
   goto try_again;
 ret= 0;
ex:;
 Sregex_string(&input_line,NULL,0);
 return(ret);
}


