
/*
  cc -g -c imgexempt.c
*/

#include <unistd.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>



#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"
#include "../s_tools/sregex.h"

#include "imgexempt.h"


/* ----------------------------- ImgexempT ------------------------------ */


int Imgexempt_new(exempt,filename,lookalikes,flag)
struct ImgexempT **exempt;
char *filename;
struct LstrinG **lookalikes;
int flag;
{
 struct ImgexempT *e;

 *exempt= e= TSOB_FELD(struct ImgexempT,1);
 if(e==NULL)
   return(-1);
 e->filename= NULL;
 e->lookalikes= NULL;
 if(Sregex_string(&(e->filename),filename,0)<=0)
   goto failed;
 e->lookalikes= *lookalikes;
 *lookalikes= NULL;            /* owner is the new exempt object now */
 return(1);
failed:
 Imgexempt_destroy(exempt,0);
 return(-1);
}


int Imgexempt_destroy(exempt,flag)
struct ImgexempT **exempt;
int flag;
{
 if((*exempt)==NULL)
   return(0);
   
 Sregex_string(&((*exempt)->filename),NULL,0);
 Lstring_destroy_all(&((*exempt)->lookalikes),0);
 Smem_freE((char *) *exempt);
 *exempt= NULL;
 return(1);
}


int Imgexempt_cmp_qsort(left,right)
/* to be used by qsort() */
struct ImgexempT **left,**right;
{
 return(strcmp((*left)->filename,(*right)->filename));
}


/* --------------------------- ImgexemptlisT ------------------------------ */


int Imgexemptlist_new(exemptlist,listname,flag)
struct ImgexemptlisT **exemptlist;
char *listname;
int flag;
{
 int ret,i;
 struct ImgexemptlisT *e;

 *exemptlist= e= TSOB_FELD(struct ImgexemptlisT,1);
 if(e==NULL)
   return(-1);
 e->refcount= 1;
 e->exempt= NULL;
 e->count= 0;
 for(i=0;i<Imgexempt_unsorted_addsizE;i++)
   e->added_exempts[i]= NULL;
 e->added_counter= 0;
 e->listname= NULL;
 e->list_mtime= 0;

 ret= Imgexemptlist_read(e,listname,0);
 if(ret<=0)
   goto failed;

 return(1);
failed:;
 Imgexemptlist_destroy(exemptlist,0);
 return(ret);
}


int Imgexemptlist_destroy(exemptlist,flag)
struct ImgexemptlisT **exemptlist;
int flag;
{
 int i;
 struct ImgexemptlisT *e;

 e= *exemptlist;
 if((e)==NULL)
   return(0);
 e->refcount--;
 if(e->refcount>0)
   return(2);
 Imgexemptlist_destroy_exempts(e,0);
 Sregex_string(&(e->listname),NULL,0);
 Smem_freE((char *) e);
 *exemptlist= NULL;
 return(1);
}


int Imgexemptlist_destroy_exempts(exemptlist,flag)
struct ImgexemptlisT *exemptlist;
int flag;
{
 int i;

 for(i=0;i<exemptlist->added_counter;i++)
   Imgexempt_destroy(&(exemptlist->added_exempts[i]),0);
 exemptlist->added_counter= 0;
 if(exemptlist->exempt!=NULL) {
   for(i=0;i<exemptlist->count;i++)
     Imgexempt_destroy(&(exemptlist->exempt[i]),0);
   Smem_freE((char *) exemptlist->exempt);
   exemptlist->exempt= NULL;
 }
 exemptlist->count= 0;
 return(1);
}


int Imgexemptlist_refer(exemptlist,referer,flag)
struct ImgexemptlisT *exemptlist;
struct ImgexemptlisT **referer;
int flag;
{
 if(exemptlist==NULL)
   return(0);
 exemptlist->refcount++;
 *referer= exemptlist;
 return(1);
}


int Imgexemptlist_read(exemptlist,listname,flag)
struct ImgexemptlisT *exemptlist;
char *listname;
int flag;
/*
 bit0= only read if listname or modification time have changed
*/
{
 int ret,count,i,wpt,state;
 FILE *fp= NULL;
 struct stat stbuf;
 char line[SfileadrL],filename[SfileadrL];
 struct LstrinG *lstring= NULL;

 /* get file information */
 if(stat(listname,&stbuf)==-1)
   return(0);
 if(flag&1) {
   /* is the list unchanged ? */
   if(exemptlist->listname!=NULL)
     if(strcmp(exemptlist->listname,listname)==0)
       if(exemptlist->list_mtime==stbuf.st_mtime)
         return(2);
 }
 Sregex_string(&(exemptlist->listname),NULL,0);
 exemptlist->list_mtime= 0;
 Imgexemptlist_destroy_exempts(exemptlist,0);
 
 count= Sfile_line_count(listname,2);
 if(count<0)
   return(0);
 count++;
 exemptlist->exempt= TSOB_FELD(struct ImgexempT *,count);
 if(exemptlist->exempt==NULL)
   return(-1);
 for(i=0;i<count;i++) 
   exemptlist->exempt[i]= NULL;

 fp= fopen(listname,"r");
 if(fp==NULL)
   return(0);
 state= 0;
 wpt= 0;
 while(Sfile_fgets(line,sizeof(line),fp)!=NULL) {
   if(line[0]==0) {
     if( state!=0) {
       ret= Imgexempt_new(&(exemptlist->exempt[wpt]),filename,&lstring,0);
       if(ret<=0)
         goto ex;
       state= 0;
       wpt++;
       if(wpt>=count)
 break;
     }
   } else if(state==0) {                            /* first line of a group */
     strcpy(filename,line);
     state= 1;
   } else {                                                /* following line */
     ret= Lstring_new(&lstring,line,lstring,1);
     if(ret<=0)
       goto ex;
   }
 }
 exemptlist->count= wpt;
 if(Sregex_string(&(exemptlist->listname),listname,0)<=0)
   {ret= -1; goto ex;}
 exemptlist->list_mtime= stbuf.st_mtime;
 ret= Imgexemptlist_sort(exemptlist,0);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 if(fp!=NULL)
   fclose(fp);
 return(ret);
}


int Imgexemptlist_sort(exemptlist,flag)
struct ImgexemptlisT *exemptlist;
int flag;
{
 if(exemptlist->count<=0)
   return(2);
 qsort(exemptlist->exempt,exemptlist->count,sizeof(struct ImgexempT *),
       Imgexempt_cmp_qsort);
 return(1);
}


int Imgexemptlist_search_idx(exemptlist,name,result_idx,flag)
struct ImgexemptlisT *exemptlist;
char *name;
int *result_idx;
int flag;
{
 int ret,lb,ub,step,idx,cmp;

 if(exemptlist->count<=0) {
   *result_idx= 0;
   return(0);
 }
 lb= 0;
 ub= exemptlist->count-1;

 while(1) {
   step= ub-lb; 
   idx= lb+step/2+step%2;
   cmp= strcmp(exemptlist->exempt[idx]->filename,name);
   if(cmp<0) {
     lb= idx;
   } else if(cmp>0) {
     if(ub==idx)
       ub= lb;
     else
       ub= idx;
   } else {
     *result_idx= idx;
     return(1);
   }
   if(step==0)
 break;
 }
 *result_idx= idx;
 if(strcmp(exemptlist->exempt[idx]->filename,name)<0)
   *result_idx++;
 return(0);
}


int Imgexemptlist_is_exempted(exemptlist,name1,name2,flag)
struct ImgexemptlisT *exemptlist;
char *name1,*name2;
int flag;
{
 int ret,idx;

 if(exemptlist==NULL)
   return(0);
 ret= Imgexemptlist_search_idx(exemptlist,name1,&idx,0);
 if(ret<0)
   return(ret);
 if(ret==0)
   goto not_in_exempt;

 /* look for first of possibly multiple exempts for name1 */
 while(idx>0) {
   if(strcmp(exemptlist->exempt[idx-1]->filename,name1)!=0)
 break;
   idx--;
 }
 /* check all exempts of name1 for existence of name2 */
 while(strcmp(exemptlist->exempt[idx]->filename,name1)==0) {
   if(Lstring_is_listed(exemptlist->exempt[idx]->lookalikes,name2,0)>0)
     return(1);
   if(idx<exemptlist->count-1)
     idx++;
   else
 break;
 }
not_in_exempt:;
 /* test wether exempt was recently added */
 for(idx=0;idx<exemptlist->added_counter;idx++)
   if(strcmp(exemptlist->added_exempts[idx]->filename,name1)==0)
     if(Lstring_is_listed(exemptlist->added_exempts[idx]->lookalikes,name2,0)>0)
       return(1);
 return(0);
}


int Imgexemptlist_can_still_add(exemptlist,flag)
struct ImgexemptlisT *exemptlist;
int flag;
{
 return(exemptlist->added_counter<Imgexempt_unsorted_addsizE);
}


int Imgexemptlist_add(exemptlist,listname,filename,lookalikes,flag)
struct ImgexemptlisT *exemptlist;
char *listname;
char *filename;
struct LstrinG **lookalikes;
int flag;
/*
 bit0= do not copy modification time from file
*/
{
 int ret;
 struct ImgexempT *exempt;
 struct stat stbuf;

 if(!Imgexemptlist_can_still_add(exemptlist,0))
   return(0);
 ret= Imgexempt_new(&exempt,filename,lookalikes,0);
 if(ret<=0)
   goto ex;
 exemptlist->added_exempts[exemptlist->added_counter]= exempt;
 exemptlist->added_counter++;
 if(flag&1)
   {ret= 1; goto ex;} 
 if(stat(listname,&stbuf)==-1)
   {ret= 0; goto ex;}
 exemptlist->list_mtime= stbuf.st_mtime;
 ret= 1;
ex:;
 return(ret);
}
