
/*
 cc -g -c imvfiles.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "imvhistory.h"
#include "imvfiles.h"


/* -------------------------------- ImvfilE ------------------------------ */


int Imvfile_new(file,adr,flag)
struct ImvfilE **file;
char *adr;
int flag;
{
 struct ImvfilE *f;
 *file= f= TSOB_FELD(struct ImvfilE,1);
 if(f==NULL)
   return(-1);
 f->valid= 1;
 f->adr= NULL;
 f->original_adr= NULL;
 f->adr= TSOB_FELD(char,strlen(adr)+1);
 if(f->adr==NULL) {
   Imvfile_destroy(file,0);
   return(-1);
 }
 strcpy(f->adr,adr);
 f->original_adr= f->adr;
 return(1);
}


int Imvfile_destroy(file,flag)
struct ImvfilE **file;
int flag;
/*
 bit0= only delete components, do not destroy **file
*/
{
 if((*file)->original_adr!=(*file)->adr) {
   if((*file)->original_adr!=NULL)
     Smem_freE((char *) (*file)->original_adr);
 }
 if((*file)->adr!=NULL)
   Smem_freE((char *) (*file)->adr);
 Smem_freE((char *) *file);
 *file= NULL;
 return(1);
}


int Imvfile_set_invalid(file,flag)
struct ImvfilE *file;
int flag;
{
 file->valid= 0;
 return(1);
}


int Imvfile_is_valid(file,flag)
struct ImvfilE *file;
int flag;
{
 return(!!file->valid);
}


int Imvfile_set_adr(file,adr,flag)
struct ImvfilE *file;
char *adr;
int flag;
{
 char *cpt;

 cpt= TSOB_FELD(char,strlen(adr)+1);
 if(cpt==NULL)
   return(-1);
 if(file->adr!=NULL && file->adr!=file->original_adr)
   Smem_freE((char *) file->adr);
 file->adr= cpt;
 strcpy(cpt,adr);
 return(1);
}


int Imvfile_get_adr(file,adr,flag)
struct ImvfilE *file;
char *adr;
int flag;
{
 int ret;

 strcpy(adr,file->adr);
 return(1);
}


/* -------------------------------- ImvfileS ------------------------------ */


int Imvfiles_new(files,agent,size,flag)
struct ImvfileS **files;
struct AsagenT *agent;
int size;
int flag;
{
 struct ImvfileS *f;
 int i;

 *files= f= TSOB_FELD(struct ImvfileS,1);
 if(f==NULL)
   return(-1);
 f->agent= agent;
 f->file= NULL;
 if(size<=0)
   size= 16;
 f->size= size;
 f->max_step= 16*size;
 f->count= 0;
 f->idx= 0;
 f->history_oldest= NULL;
 f->history_youngest= NULL;
 f->file= TSOB_FELD(struct ImvfilE *,size);
 if(f->file==NULL) {
   Imvfiles_destroy(files,0);
   return(-1);
 }
 for(i=0;i<size;i++)
   f->file[i]= NULL;
 return(1);
}


int Imvfiles_destroy(files,flag)
struct ImvfileS **files;
int flag;
{
 int i;
 struct ImvfilE **file;
 struct ImvhistorY *curr_h,*next_h;

 if(*files==NULL)
   return(0);
 curr_h= (*files)->history_youngest;
 while(curr_h!=NULL) {
   next_h= curr_h->prev;
   Imvhistory_destroy(&(curr_h),0);
   curr_h= next_h;
 }
 if((*files)->file!=NULL) {
   file= (*files)->file;
   for(i=(*files)->count-1;i>=0;i--)
     if(file[i]!=NULL) 
       Imvfile_destroy(&(file[i]),0);
   Smem_freE((char *) file);
 }
 Smem_freE((char *) (*files));
 *files= NULL;
 return(1);
}  


int Imvfiles_return_count(files,flag)
struct ImvfileS *files;
int flag;
{
 return(files->count);
}


int Imvfiles_get_idx(files,idx,flag)
struct ImvfileS *files;
int *idx;
int flag;
{
 *idx= -1;
 if(files->count<=0)
   return(0);
 *idx= files->idx;
 return(1);
}


int Imvfiles_set_idx(files,idx,flag)
struct ImvfileS *files;
int idx;
int flag;
/*
 bit0= add to current index (but skip invalid entries)
*/
{
 int ret= 1,i;

 if(flag&1) {
   if(files->idx<0) 
     return(0);
   if(idx>0) {
     for(i= files->idx;i<files->count && idx>0;i++)
       if(Imvfile_is_valid(files->file[i],0))
         idx--;
   } else {
     for(i= files->idx;i>=0 && idx<0;i--)
       if(Imvfile_is_valid(files->file[i],0))
         idx++;
   }
   idx= i;
 }
 if(idx<0){
   idx= 0;
   ret= 0;
 }
 if(idx>=files->count) {
   idx= files->count-1;
   ret= 0;
 }
 files->idx= idx;
 return(ret);
}


int Imvfiles_get_idxadr(files,idx,adr,flag)
struct ImvfileS *files;
int idx;
char *adr;
int flag;
{
 int ret;

 if(files->count<=0)
   return(0);
 if(idx<0||idx>=files->count)
   return(0);
 ret= Imvfile_get_adr(files->file[idx],adr,0);
 return(ret);
}


int Imvfiles_get_adr(files,adr,flag)
struct ImvfileS *files;
char *adr;
int flag;
{
 int ret;

 ret= Imvfiles_get_idxadr(files,files->idx,adr,0);
 return(ret);
}


int Imvfiles_set_adr(files,adr,flag)
struct ImvfileS *files;
char *adr;
int flag;
{
 int ret,different;

 if(files->count<=0)
   return(0);
 different= strcmp(files->file[files->idx]->adr,adr);
 if(!different)
   return(2);
 ret= Imvfile_set_adr(files->file[files->idx],adr,0);
 if(ret<=0)
   return(ret);
 Asagent_notify_list_change(files->agent,files->idx,
                             "change",files->file[files->idx]->adr,0);
 return(1);
}


int Imvfiles_set_invalid(files,flag)
struct ImvfileS *files;
int flag;
{
 Imvfile_set_invalid(files->file[files->idx],0);
 return(1);
}


int Imvfiles_enlarge(files,step,flag)
struct ImvfileS *files;
int step;
int flag;
{
 struct ImvfilE **file;
 int i,size;

 if(step<=0){
   step= files->size*2;
   if(step>files->max_step)
     step= files->max_step;
 }
 size= files->size+step;
 file= TSOB_FELD(struct ImvfilE *,size);
 if(file==NULL) 
   return(-1);
 for(i=0;i<files->count;i++)
   file[i]= files->file[i];
 for(i=files->count;i<size;i++)
   file[i]= NULL;
 Smem_freE((char *) files->file);
 files->file= file;
 files->size= size;
 return(1);
}


int Imvfiles_is_bad_name(files,eff_adr,flag)
struct ImvfileS *files;
char *eff_adr;
int flag;
/*
 return:
  0= legal name
  1= too long
  2= illegal characters
*/
{
 int i,l;

 l= strlen(eff_adr);
 if(l>SfileadrL/2)            /* should be enough for any reasonable fileadr */
   return(1);                 /* and save fixed length strings from problems */
 if(strchr(eff_adr,'$')!=NULL)
   return(2);
 if(strchr(eff_adr,'`')!=NULL)
   return(2);
 if(strchr(eff_adr,'\\')!=NULL)
   return(2);
 if(strchr(eff_adr,'(')!=NULL)
   return(2);
 if(strchr(eff_adr,')')!=NULL)
   return(2);
 if(strchr(eff_adr,';')!=NULL)
   return(2);
 for(i=0;i<l;i++)
   if(isspace(eff_adr[i]))
     return(2);
 return(0);
}


int Imvfiles_add(files,adr,flag)
struct ImvfileS *files;
char *adr;
int flag;
/*
 bit0= set index to new idem
 bit1= reuse last current working directory (if set)
return:
 <=0 = lower level error
   1 = ok
   2 = illegal filename refused
*/
{
 int ret;
 char *eff_adr= NULL,*dir= NULL;
 static char current_working_dir[SfileadrL]={""};
 struct stat stbuf;

 if(stat(adr,&stbuf)==-1)
   {ret= 3; goto ex;}
 if(!S_ISREG(stbuf.st_mode))
   {ret= 3; goto ex;}

 if(files->count>=files->size) {
   ret= Imvfiles_enlarge(files,0,0);
   if(ret<=0)
     goto ex;
 }
 if(adr[0]!='/') {
   if((flag&2) && current_working_dir[0]!=0) {
     if(Sregex_string(&eff_adr,current_working_dir,0)<=0)
       {ret= -1; goto ex;}
   }else{
     ret= Sfile_getcwd(&dir,1);
     if(ret>0) {
       strcpy(current_working_dir,dir);
       ret= Sregex_string(&eff_adr,dir,0);
       if(ret<=0)
         {ret= -1; goto ex;}
     }
   }
 }
 if(Sregex_string(&eff_adr,adr,1)<=0)
   {ret= -1; goto ex;}
 if(Imvfiles_is_bad_name(files,eff_adr,0))
   {ret= 2; goto ex;}
 ret= Imvfile_new(&(files->file[files->count]),eff_adr,0);
 if(ret<=0)
   goto ex;
 if(flag&1)
   files->idx= files->count;
 files->count++;
 Asagent_notify_list_change(files->agent,files->idx,"append",eff_adr,0);
 ret= 1;
ex:;
 Sregex_string(&dir,NULL,0);
 Sregex_string(&eff_adr,NULL,0);
 return(ret);
}


int Imvfiles_add_history(files,history,flag)
struct ImvfileS *files;
struct ImvhistorY *history;
int flag;
{
 if(files->history_oldest==NULL)
   files->history_oldest= history;
 if(files->history_youngest==NULL) {
   history->prev= NULL;
 } else {
   history->prev= files->history_youngest;
   files->history_youngest->next= history;
 }
 files->history_youngest= history;
 history->next= NULL;
 return(1);
}


int Imvfiles_pop_history(files,history,flag)
struct ImvfileS *files;
struct ImvhistorY **history;
int flag;
{
 *history= NULL;
 if(files->history_youngest==NULL)
   return(0);
 *history= files->history_youngest;
 files->history_youngest= files->history_youngest->prev;
 if(files->history_youngest!=NULL)
   files->history_youngest->next= NULL;
 else
   files->history_oldest= NULL;
 (*history)->prev= NULL;
 (*history)->next= NULL;
 return(1);
}

