
/*
 cc -g -c imvhistory.c
*/

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>


#include "../s_tools/smem.h"
#include "../s_tools/sfile.h"

#include "../as/asevent.h"
#include "imvhistory.h"


/* -------------------------------- ImvhistorY ------------------------------ */


int Imvhistory_new(hist,idx,source_adr,current_adr,replaced_by_adr,event,flag)
struct ImvhistorY **hist;
int idx;
char *source_adr,*current_adr,*replaced_by_adr;
struct AsevenT *event;
int flag;
/*
 bit1= checked_in
 bit2= checked_out
 bit3= replaced by a descendent of itself
 bit4= no undo possible with this operation
 bit5= nothing has happened in this operation
*/
{
 struct ImvhistorY *h;
 struct timezone tz;
 int ret;

 *hist= h= TSOB_FELD(struct ImvhistorY,1);
 if(h==NULL)
   return(-1);
 h->idx= idx;
 h->source_adr= NULL;
 h->current_adr= NULL;
 h->replaced_by_adr= NULL;
 h->event= event;
 h->checked_in= !!(flag&2);
 h->checked_out= !!(flag&4); 
 h->replaced_by_itself= !!(flag&8);
 h->no_undo_possible= !!(flag&16);
 h->nothing_happened= !!(flag&32);
 gettimeofday(&(h->date),&tz);
 h->next= h->prev= NULL;
 h->boss= h->sub= NULL;
 ret= Imvhistory_set_source(h,source_adr,0);
 if(ret<=0) {
failed:;
   Imvhistory_destroy(hist,0);
   return(ret);
 }
 ret= Imvhistory_set_current(h,current_adr,0);
 if(ret<=0) 
   goto failed;
 ret= Imvhistory_set_replaced_by(h,replaced_by_adr,0);
 if(ret<=0) 
   goto failed;
 ret= Asevent_new(&(h->event),"",-1,"",0);
 if(ret<=0) 
   goto failed;
 ret= Asevent_copy(event,h->event,0);
 if(ret<=0) 
   goto failed;
 return(1);
}


int Imvhistory_destroy(hist,flag)
struct ImvhistorY **hist;
int flag;
{
 struct ImvhistorY *h,*sub,*subnext;
 int hist_is_boss_sub= 0;

 h= *hist;
 if(h==NULL)
   return(0);

 /* unlink and eventually relink connections */
 if(h->boss!=NULL) {
   if(hist==&(h->boss->sub))
     hist_is_boss_sub= 1;              /* protects *hist from being NULLed */
   h->boss->sub= NULL;
   if(h->next!=NULL) {
     if(h->next->boss==NULL) {
       h->boss->sub= h->next;
       h->next->boss= h->boss;
     }
   }
 }
 if(h->prev!=NULL)
   h->prev->next= h->next;
 if(h->next!=NULL)
   h->next->prev= h->prev;
 
 /* destroy subtree */
 sub= h->sub;
 while(sub!=NULL) {
   subnext= sub->next;
   Imvhistory_destroy(&sub,0);
   sub= subnext;
 }

 /* destroy own storage */
 if(h->source_adr!=NULL)
   Smem_freE((char *) h->source_adr);
 if(h->current_adr!=NULL)
   Smem_freE((char *) h->current_adr);
 if(h->replaced_by_adr!=NULL)
   Smem_freE((char *) h->replaced_by_adr);
 Asevent_destroy(&(h->event),0);  
 Smem_freE((char *) h);
 if(!hist_is_boss_sub)   /* else : hist was adjusted while relinking */
   *hist= NULL;
 return(1);
}


int Imvhistory_set_idx(hist,idx,flag)
struct ImvhistorY *hist;
int idx;
int flag;
{
 if(hist==NULL)
   return(0);
 hist->idx= idx;
 return(1);
}


int Imvhistory_get_idx(hist,idx,flag)
struct ImvhistorY *hist;
int *idx;
int flag;
{
 if(hist==NULL)
   return(0);
 *idx= hist->idx;
 return(1);
}


int Imvhistory_set_source(hist,adr,flag)
struct ImvhistorY *hist;
char *adr;
int flag;
{
 int ret;
 if(hist==NULL)
   return(0);
 ret= Sregex_string(&(hist->source_adr),adr,0);
 return(ret);
}


int Imvhistory_get_source(hist,adr,flag)
struct ImvhistorY *hist;
char **adr;
int flag;
{
 int ret;
 if(hist==NULL)
   return(0);
 ret= Sregex_string(adr,hist->source_adr,0);
 return(ret);
}


int Imvhistory_set_current(hist,adr,flag)
struct ImvhistorY *hist;
char *adr;
int flag;
{
 int ret;
 if(hist==NULL)
   return(0);
 ret= Sregex_string(&(hist->current_adr),adr,0);
 return(ret);
}


int Imvhistory_get_current(hist,adr,flag)
struct ImvhistorY *hist;
char **adr;
int flag;
{
 int ret;
 if(hist==NULL)
   return(0);
 ret= Sregex_string(adr,hist->current_adr,0);
 return(ret);
}


int Imvhistory_set_replaced_by(hist,adr,flag)
struct ImvhistorY *hist;
char *adr;
int flag;
/*
 bit3= replaced by a descendent of itself
*/
{
 int ret;
 if(hist==NULL)
   return(0);
 ret= Sregex_string(&(hist->replaced_by_adr),adr,0);
 if(flag&8)
   hist->replaced_by_itself= 1;
 return(ret);
}


int Imvhistory_replaced_by_itself(hist,flag)
struct ImvhistorY *hist;
int flag;
{
 if(hist==NULL)
   return(0);
 return(hist->replaced_by_itself);
}


int Imvhistory_set_no_undo_possible(hist,value,flag)
struct ImvhistorY *hist;
int value;
int flag;
{
 if(hist==NULL)
   return(2);
 hist->no_undo_possible= !!value;
 return(1);
}


int Imvhistory_no_undo_possible(hist,flag)
struct ImvhistorY *hist;
int flag;
{
 if(hist==NULL)
   return(2);
 return(hist->no_undo_possible);
}


int Imvhistory_set_nothing_happened(hist,value,flag)
struct ImvhistorY *hist;
int value;
int flag;
{
 if(hist==NULL)
   return(2);
 hist->nothing_happened= !!value;
 return(1);
}


int Imvhistory_nothing_happened(hist,flag)
struct ImvhistorY *hist;
int flag;
{
 if(hist==NULL)
   return(2);
 return(hist->nothing_happened);
}


int Imvhistory_link_sub(boss,sub,flag)
struct ImvhistorY *boss,*sub;
int flag;
{
 struct ImvhistorY *boss_old_sub= NULL,*sub_old_boss= NULL,*sub_last_sub= NULL;

 if(boss!=NULL)
   boss_old_sub= boss->sub;
 if(sub!=NULL)
   sub_old_boss= sub->boss;
 if(boss!=NULL) {
   if(boss_old_sub!=NULL) {
     boss_old_sub->boss= NULL;
     if(sub!=NULL) {
       sub_last_sub= sub;
       while(1)
         if(sub_last_sub->sub==NULL)
       break;
         else
           sub_last_sub= sub_last_sub->sub;
       boss_old_sub->boss= sub_last_sub;
       sub_last_sub->sub= boss_old_sub; 
     }
   } 
   boss->sub= sub;
 }
 if(sub!=NULL) {
   if(sub_old_boss!=NULL)
     sub_old_boss->sub= NULL;
   sub->boss= boss;
 }
 return(1);
}


int Imvhistory_link_next(prev,next,flag)
struct ImvhistorY *prev,*next;
int flag;
{
 struct ImvhistorY *prev_old_next= NULL,*next_old_prev= NULL;
 struct ImvhistorY *next_last_next= NULL;

 if(prev!=NULL)
   prev_old_next= prev->next;
 if(next!=NULL)
   next_old_prev= next->prev;
 if(prev!=NULL) {
   if(prev_old_next!=NULL) {
     prev_old_next->prev= NULL;
     if(next!=NULL) {
       next_last_next= next;
       while(1)
         if(next_last_next->next==NULL)
       break;
         else
           next_last_next= next_last_next->next;
       prev_old_next->prev= next_last_next;
       next_last_next->next= prev_old_next; 
     }
   } 
   prev->next= next;
 }
 if(next!=NULL) {
   if(next_old_prev!=NULL)
     next_old_prev->next= NULL;
   next->prev= prev;
 }
 return(1);
}



