
/*
  cc -g -c nntppost.c
*/

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>


#define TSOB_FELD(typ,anz) (typ *) malloc((anz)*sizeof(typ));

#include "../s_tools/sfile.h"

#include "nntpconn.h"
#include "nntpgroup.h"
#include "nntphead.h"
#include "nntppost.h"


/* --------------------- Nntppost -------------------- */


int Nntppost_new(postjob,group,hostid,username,organization,
                         subject,body,bin_file,flag)
struct NntpposT **postjob;
struct NntpgrouP *group;
char *hostid;
char *username;
char *organization;
char *subject;
char *body;
char *bin_file;
int flag;
/*
 bit0= body is the name of a text file rather than the body text itself
 bit1= value for use_own_msgid
*/
{
 int ret;
 struct NntpposT *p;

 *postjob= p= TSOB_FELD(struct NntpposT,1);
 if(p==NULL)
   goto ex;
 p->group= group;
 p->use_own_msgid= !!(flag&2);
 strcpy(p->hostid,hostid);
 strcpy(p->username,username);
 strcpy(p->organization,organization);
 strcpy(p->subject,subject);
 p->reference[0]= 0;
 p->all_references[0]= 0;
 p->control_msg[0]= 0;
 p->body_text= NULL;
 p->body_file[0]= 0;
 if(flag&1){
   p->body_text= TSOB_FELD(char,1);
   if(p->body_text==NULL)
     goto ex;
   p->body_text[0]= 0;
   strcpy(p->body_file,body);
 } else {
   p->body_text= TSOB_FELD(char,strlen(body)+1);
   if(p->body_text==NULL)
     goto ex;
   strcpy(p->body_text,body);
 }
 strcpy(p->bin_file,bin_file);
 p->msgid[0]= 0;
 Nntppost_unique_key(p,p->prev_key,1);
 return(1);
ex:;
 group->conn->errn= -17;
 sprintf(group->conn->errmsg,
         "Cannot create postjob object : %s\n",strerror(errno));
 if(p!=NULL)
   Nntppost_destroy(postjob,0);
 return(-1);
}


int Nntppost_destroy(postjob,flag)
struct NntpposT **postjob;
int flag;
{
 struct NntpposT *p;

 p= *postjob;
 if(p==NULL)
   return(0);
 if(p->body_text!=NULL)
   free(p->body_text);
 free((char *) p);
 *postjob= NULL;
 return(1);
}


int Nntppost_set_bin_file(postjob,filename,flag)
struct NntpposT *postjob;
char *filename;
int flag;
{
 strcpy(postjob->bin_file,filename);
 return(1);
}


int Nntppost_set_subject(postjob,text,flag)
struct NntpposT *postjob;
char *text;
int flag;
{
 strcpy(postjob->subject,text);
 return(1);
}


int Nntppost_set_all_references(postjob,text,flag)
struct NntpposT *postjob;
char *text;
int flag;
/*
 bit0= append rather than overwrite
*/
{
 int lt,la;

 lt= strlen(text);
 if(lt<=0)
   return(2);
 if(!(flag&1))
   postjob->all_references[0]= 0;
 la= strlen(postjob->all_references);
 if(la+lt>=sizeof(postjob->all_references)){
   postjob->group->conn->errn= -28;
   sprintf(postjob->group->conn->errmsg,
           "List of reference ids too long :\n%s\n+ %s",
           postjob->all_references,text);
   return(0);
 }
 if(la)
   strcat(postjob->all_references," ");
 strcat(postjob->all_references,text);
 return(1);
}

int Nntppost_set_reference(postjob,text,flag)
struct NntpposT *postjob;
char *text;
int flag;
{
 strcpy(postjob->reference,text);
 return(1);
}


int Nntppost_set_control_msg(postjob,text,flag)
struct NntpposT *postjob;
char *text;
int flag;
{
 strcpy(postjob->control_msg,text);
 return(1);
}


int Nntppost_send(postjob,line,flag)
struct NntpposT *postjob;
char *line;
int flag;
/*
 bit0= report on stdout
 bit2= do not duplicate leading dot
*/
{
 int sret,i,l;

 if(flag&1) {
   printf("%s",line);
   l= strlen(line);
   if(l>0)
     if(line[l-1]!='\n') 
       printf("\n",line);
 }
 if(!(flag&2))
   if(line[0]=='.') /* duplicate leading dot */
     for(i=strlen(line);i>=0;i--) line[i+1]= line[i];
 sret= Nntpconn_send(postjob->group->conn,line,0);
 if(sret<=0)
   return(sret);
 return(1);
}


int Nntppost_unique_key(postjob,ret_key,flag)
struct NntpposT *postjob;
char *ret_key;
int flag;
/*
 bit0= do not test with postjob->prev_key
*/
{
 struct timeval tv;
 struct timezone tz;

 /* >>> offer option to get the id from external program */

 while(1){
   gettimeofday(&tv,&tz);
   sprintf(ret_key,"%-4.4X%-8.8X%-8.8X%-4.4ld",
                   getpid(),((unsigned long) postjob)&0xFFFFFFFF,
                   tv.tv_sec,tv.tv_usec/100);
   if(flag&1)
 break;
   if(strcmp(ret_key,postjob->prev_key)==0){
     Sfile_microsleep(1000,0);
 continue;
   }
   strcpy(postjob->prev_key,ret_key);
 break;
 }
 return(1);
}


int Nntppost_head(postjob,flag)
struct NntpposT *postjob;
int flag;
{
#define NpsenD if((sret= Nntppost_send(postjob,line,1))<=0) return(sret)
 int sret,ret;
 char line[Nntppost_stringleN];
 char version[Nntppost_stringleN];
 char date[Nntppost_stringleN];
 char key[Nntppost_stringleN];

 sprintf(version,"snntpbatch-%s",Nntpversion(0));
 ret= Sfile_date(date,0);
 if(ret<=0)
   strcpy(date,"Thu, 01 Jan 1970 00:00:00");
 Nntppost_unique_key(postjob,key,0);
 sprintf(postjob->msgid,"<%s@%s>",key,postjob->hostid);

/*
 sprintf(line,"Relay-Version: version %s",version); NpsenD;
 sprintf(line,"Posting-Version: version %s",version); NpsenD;
*/
 sprintf(line,"User-Agent: %s",version); NpsenD;
 if(strchr(postjob->username,'@')!=NULL)
   sprintf(line,"From: %s",postjob->username);
 else 
   sprintf(line,"From: %s@%s",postjob->username,postjob->hostid);
 NpsenD;
 sprintf(line,"Date: %s",date); NpsenD;
 sprintf(line,"Newsgroups: %s",postjob->group->name); NpsenD;
 sprintf(line,"Subject: %s",postjob->subject); NpsenD;
 if(postjob->use_own_msgid) {
   sprintf(line,"Message-ID: %s",postjob->msgid); NpsenD;
 }
 sprintf(line,"Path: %s",postjob->group->conn->servername); NpsenD;
 sprintf(line,"Organization: %s",postjob->organization); NpsenD;
 if(postjob->all_references[0]!=0){
    sprintf(line,"References: %s",postjob->all_references); NpsenD;
 }
 if(postjob->control_msg[0]!=0){
    sprintf(line,"Control: %s",postjob->control_msg); NpsenD;
 }
 /* header end mark is an empty line */
 line[0]= 0; NpsenD;
 return(1);
#undef NpsenD
}


int Nntppost_msg_text(postjob,flag)
struct NntpposT *postjob;
int flag;
{
 int sret,ret;
 char line[Nntppost_stringleN],*spt,*cpt;
 FILE *fp;
 struct NntpconN *conn;
 conn= postjob->group->conn;

 /* body_text */
 if(postjob->body_text!=NULL)
 if(postjob->body_text[0]) {
   cpt= postjob->body_text;
   while((*cpt)!=0){
     spt= strchr(cpt,'\n');
     if(spt==NULL)
       spt= cpt+strlen(cpt);
     strncpy(line,cpt,spt-cpt);
     line[spt-cpt]= 0;
     sret= Nntppost_send(postjob,line,1);
     if(sret<=0)
       return(sret);
     if((*spt)==0)
   break;
     cpt= spt+1;
   }
 }

 /* body_file */
 if(postjob->body_file[0]!=0){
   fp= Sfile_fopen(postjob->body_file,"r",0);
   if(fp==NULL){
     conn->errn= errno;
     sprintf(conn->errmsg,"Error %ld on opening input file : %s\n%s",
                          errno,postjob->body_file,strerror(errno));
     return(0);
   }
   while(fgets(line,sizeof(line),fp)!=NULL){
     sret= Nntppost_send(postjob,line,1);
     if(sret<=0)
       return(sret);
   }
   Sfile_fclose(fp);
 }

 return(1);
}


int Nntppost_bin_file(postjob,flag)
struct NntpposT *postjob;
int flag;
{
 int sret,ret;
 char line[Nntppost_stringleN],*spt,*cpt;
 char cmd[SfileadrL];
 int lin_count;
 FILE *fp;

 if(postjob->bin_file[0]==0)
   return(2);
 cpt= strrchr(postjob->bin_file,'/');
 if(cpt==NULL)
   cpt= postjob->bin_file;
 else
   cpt++;
 sprintf(cmd,Nntppost_uuencodeR,cpt,postjob->bin_file);
 printf("%s\n",cmd);
 fp= popen(cmd,"r");
 if(fp==NULL){
   postjob->group->conn->errn= errno;
   sprintf(postjob->group->conn->errmsg,
           "Error %ld on opening pipe to : %s\n%s",errno,cmd,strerror(errno));
   return(-1);
 }
 /* read from pipe until EOF */
 lin_count= 0;
 while(Sfile_fgets(line,sizeof(line)-10,fp)!=NULL){
   ret= Nntppost_send(postjob,line,0);
   if(ret<=0)
     goto ex;
   lin_count++;
   Nntpconn_enforce_limits(postjob->group->conn,
                           postjob->group->conn->baudlimitfile,
                           ((lin_count%100)==0)|4);
   if(!postjob->group->conn->show_io_stat)
     fprintf(stderr,"\r  %ld uuencoded lines\r",lin_count);
 }
 if(!postjob->group->conn->show_io_stat)
   Nntpconn_show_stat(postjob->group->conn,1); /* clear line */
 fprintf(stderr,"\r  %ld uuencoded lines posted\n",lin_count);
 ret= 1;
ex:;
 if(fp!=NULL) pclose(fp);
 return(ret);
}


int Nntppost_make_reflist(postjob,flag)
struct NntpposT *postjob;
int flag;
{
 struct NntpheaD *head;
 int ret,l;

 postjob->all_references[0]= 0;
 if(postjob->reference[0]==0)
   return(2);
 ret= Nntphead_new(&head,postjob->group->conn,postjob->reference,0);
 if(ret<=0)
   goto ex;
 ret= Nntppost_set_all_references(postjob,head->all_references,0);
 if(ret<=0)
   goto ex;
 ret= Nntppost_set_all_references(postjob,postjob->reference,1);
 if(ret<=0)
   goto ex;
 ret= 1;
ex:;
 Nntphead_destroy(&head,0);
 return(ret);
}


int Nntppost_article(postjob,flag)
struct NntpposT *postjob;
int flag;
/*
 bit0= postjob->all_references is already set properly.
*/
{
 struct NntpconN *conn;
 int poststat,ret;
 char text[81];
 struct stat stbuf;

 conn= postjob->group->conn;

 if(postjob->bin_file[0]!=0)
   if(stat(postjob->bin_file,&stbuf)==-1){
     conn->errn= ENOENT;
     sprintf(conn->errmsg,"missing binary file : %s",postjob->bin_file); 
     ret= 0; goto ex;
   }

 if(!(flag&1)){
   ret= Nntppost_make_reflist(postjob,0);
   if(ret<=0)
     goto ex;
 }
 poststat= Nntpconn_transaction(conn,"POST",conn->buffer,
                                sizeof(conn->buffer),"",2);
 if(poststat<=0)
   {ret= poststat; goto ex;}
 if(conn->status_code!=340){
   conn->errn= -25;
   sprintf(conn->errmsg,"Server does not permit posting in group '%s'",
                        postjob->group->name);
   ret= 0; goto ex;
 }

 ret= Nntppost_head(postjob,0);
 if(ret<=0)
   goto ex;
 ret= Nntppost_msg_text(postjob,0);
 if(ret<=0)
   goto ex;
 ret= Nntppost_bin_file(postjob,0);
 if(ret<=0)
   goto ex;

 /* close message body */
 strcpy(text,".");
 ret= Nntppost_send(postjob,text,1|2);
 if(ret<=0)
   goto ex;
 printf("message sent. waiting for confirmation ...\n");
 ret= Nntpconn_recv_status(conn,1);
 if(ret<=0)
   return(ret);
 Nntpconn_show_stat(conn,1); /* clear line */
 if(conn->status_code==240)
   printf("... accepted.\n");
 else
   printf("... rejected.\nstatus = %s\n",conn->status_line);

 ret= 1;
ex:;
 return(ret);
}


