

#ifndef Saccess_includeD
#define Saccess_includeD Yes



struct SaccessiteM {

 char *topic;            
 char *target;
 char *property;    
 char *value;

 struct SaccessiteM *next,*prev;
};


struct SaccesS {

 struct SaccessiteM *itemlist;
 int itemcount;
  
};


/*

A SaccesS is a simple storage and retrieval system for tuples of four texts:
  topic , target , option , value
To the SaccesS object, only the texts "all" and "any" have a meaning.
Inside the value component, even they are texts like any other. 

SaccesS is used for storing access control rules and for interchanging
messages between very loosely coupled program modules. More applications
might be added in future.

A SaccesS object is a list of SaccessiteM objects. New items are added by
Saccess_append() . There are two ways to retrieve such items: 
Saccess_find_value() returns a pointer to the value of the youngest matching
item. Saccess_consume_last() returns a copy of the value of the oldest matching
item and deletes this item from the list.

Wether an item matches a query is determined by comparing topic, target and
option with three search texts. If an item's component contains the text "all"
then it matches any given search text. If a search text contains the text "any"
it matches any given item's component.
Any other component's content has to be exactly equal to the corresponding
search text in order to match.

Examples:

The search 
 {"connection","tcp","tunnel"}
matches
 {"connection","tcp","tunnel","allow"}
 {"connection","all","tunnel","refuse"}
but not
 {"connection","file","tunnel","refuse"}

The search
 {"message","stderr","any"}
matches 
 {"message","stderr","tunnel","...some message..."}
but not
 {"message","stdout","tunnel","...some message..."}

------------------------------------------------------------------------------

Overview on used combinations :

 topic="head"                                        (dummy item at list start)
   target="head"
     option="head"
       value=""
 topic="connection"
   target={"none","tcp","file","internal","all","<"+filename}
     option="tunnel"
       value={"require","allow","refuse"}
     option="controller"
       value={"on","off"}
     option="keydirectory"
       value=filename
     option="user"
       value=username          (file keydirectory/username+".tnl" has to exist)
 topic="permission"
   target= {username,"all"}
     option= {class,class"-"action,"all"}
             class={"help","security","service","info","modify","end","shell"
                    "script","cancel","navigation","move","filelist"}
       value={"permit","deny"}
 topic="message"
   target={"stdout","stderr"}
     option=some_sender_information               (usually the fd of initiator)
       value=textline
*/

/* Examples:
   The list hook known to SaccesS:
    {"head","head","head",""} 
   Enable security control commands at file /tmp/pipe_2330
    {"connection","</tmp/pipe_2330","controller","on"}
   Require StunneL handshake and transmission for all TCP/IP connections
    {"connection","tcp","tunnel","require"}

   Deny use of shell commands (here only a class is specified)
    {"permission","all","shell","deny"}
   Allow some special shell commands for all users (here action is specified)
    {"permission","all","shell-/home/scripts/infoscript","allow"}
    {"permission","all","shell-/home/scripts/auto_move_target $*","allow"}
   Finally keep unidentified external users from executing any shell command:
    {"permission","external_anonymous","shell","deny"}

   Send a message from tunnel subsystem to the standard error facility of
   the application which may be fetch it by Saccess_consume_last().
    {"message","tunnel","stderr",
     "--- tunnel 10 : handshake reply does not match challenge."}

*/



#endif /* ! Saccess_includeD */
