

#ifndef Stunnel_includeD
#define Stunnel_includeD


/* debugging only , not intended for applications */
/* #define Stunnel_no_fishinG Yes */
/* #define Stunnel_no_formaT Yes */


/* not more than 56 , should be multiples of 8 */
#define Stunnel_keyleN 16 

/* must be multiples of 8 */
#define Stunnel_challeN 8

#define Stunnel_userleN 256

/* a token's head consists of :
  crc[4],len[4],seqno[4],rnd[4]
  seen as integers represented with high bytes first, len is signed
  each token is padded up with 0 to the next multiple of eight if len%8!=0

  len<0 indicates a control message : type[4],pad[4],data[-len]
    type= 0 : payload data (will hardly occur as control message)
          1 : reply of Stunnel_handshake_reply()
          2 : session_key of Stunnel_handshake_reply()
          3 : authentication opportunity, senseless data just to check crc 
    pad=  0 (for now)
  len>= indicates a data message ; data[len]

*/
#define Stunnel_headleN 16

/* larger tokens have to be divided automatically */
/* has to larger than Stunnel_challeN+Stunnel_userleN+Stunnel_keyleN */
#define Stunnel_max_tokenleN 1024


struct StunneL {

 /* wether there has been received any valid token */
 int authenticated;

 /* name of user identity */
 char user[Stunnel_userleN+1];
 char *keydir;

 /* opaque handle for owner of this object (filedescriptor of connection) */
 int idno;

 /* used to read and write keylists */
 struct SblowfisH *internal_fish;

 /* used to generate "random" numbers  */
 struct SblowfisH *random_fish;
 double random_count;

 /* used to authenticate client and server while session handshake */
 struct SblowfisH *personal_fish;

 /* choosen by server. used to encrypt data transmission */
 struct SblowfisH *session_fish;


 /* last challenge sent */
 char challenge[Stunnel_challeN+1];

 /* reply to be sent */
 char reply[Stunnel_challeN+1];

 /* last reply received */
 char in_reply[Stunnel_challeN+1];
 
 /* last challenge received */
 char in_challenge[Stunnel_challeN+1];


 /* status of receiving buffers :
      0=closed , 1=waiting for head , 2=waiting for payload data, 
      3=waiting for control type */
 int in_status;

 /* to assemble a token header */
 char in_head[Stunnel_headleN];
 int head_count;

 /* to assemble a complete token */
 char *data_buffer;
 int data_len;
 int data_count;
 unsigned long data_crc;


 /* sequence numbers to make tokens individual */
 unsigned long in_seqno;
 unsigned long out_seqno;

 char errmsg[SfileadrL*2];

 struct StunneL *next,*prev;

};

#define Stunnel_internal_keY "stunnel Default Configuration"


/*

A StunneL object is one half of a quite secure virtual communications
connection.
Security in this case means, that a third party may not easily read or alter
the data transmitted. A faulty data transmission will appear as such an
outside offender to the owner of StunneL.

Currently security is achieved by the blowfish encryption algorithm using
128-bit keys and by a 32-bit CRC seal. Replay of recorded transmissions is
prevented by a 32-bit sequence number which has to rise in single steps.
The key is stored in a file which is choosen by the username given to the
StunneL object and the name of its key directory.


Sending:

Users of a StunneL can put cleartext into Stunnel_format() and get encrypted
data in reply, which can be sent over insecure connections.
The result of such a conversion consists of one or more tokens assembled in
a single character array. To prevent giant buffers there is a rather small
maximum size for a single token (Stunnel_max_tokenleN).

A token consists of its head, an optional control type and a data part
of variable length. See above #define Stunnel_headleN for the exact format.
Binary 32-bit numbers are represented in a bytesex independant way.


Receiving:

The receiver of StunneL tokens may hand them over to a corresponding
StunneL object by Stunnel_retrieve() and get the original cleartext.
The decoder of StunneL will buffer data as long as a token is complete. The
content of any token completed by the input to Stunnel_retrieve() will be
concatenated and form the reply.
Be aware that Stunnel_retrieve() may have no data at all available and
return 2. If Stunnel_has_bytes()>0 , then more data is expected. Else
there has been a control message which was consumed by the StunneL object. 


Establishing a StunneL Pair:
 
Despite it is possible to create two matching StunneL objects without any
communication between their owners, usually the parameters are negotiated
at the beginning of communication. The key is depicted by a username which
leads to a file containing the key text in a locally encrypted format.
Two participants are needed for this negociation: the initiating peer and the
receiving peer. Both are using StunneL and at start are able to communicate
at least one-way from initiating peer to receiving peer. 

Initiating Peer:

Stunnel_handshake_init() produces a string of printable characters which can
be sent to the receiving peer. This string contains the username, a 64-bit
challenge encrypted by the user's key and the name of a backchannel to be
used for reply.
The backchannel may be necessary if the communication relies on oneway
connections (e.g pipes). Seen from StunneL the backchannel name is simply a
text to be transported. Seen from the using applications, it may be important
or not.

The handshake string has to be transmitted to the receiving peer now.

The reply coming back from the peer is handed over to Stunnel_handshake_accept()
which will return 0 if the authentication failed, 1 if the StunneL pair is
ready for transmissions or 2 if more reply data are needed. Excess reply
data is marked for use by the application. So if the application protocol
allows the peer to send immediately, be sure to check that pointer.

See Sconn_try_tunnel() for an example.

Receiving Peer:

The receiving peer should call Stunnel_handshake_reply() which tries to obtain
the user's key from file. If the key is available, the challenge gets its reply
and a key for the current session is choosen. Both results are formatted as
tokens by the user's key. The session key will be used for all further
transmissions.
Also the backchannel name is retrieved from the handshake string. If the
application (e.g. SconN) does not have established a sending channel yet, it
can use this name for doing so.
The two reply token have to be sent to the initiating peer now.

See Sconn_expect_tunnel() for an example.

After Handshake:

Now the handshake is complete. The receiving peer has authenticated via the
64-bit reply to the 64-bit challenge. The initiating peer will authenticate by
having understood the session key. Without that key, it will not be able
to format a valid token.

So if the receiving peer has to be sure about the authenticity of the initiating
peer, the application protocol should demand some initiating message before 
normal communications start. Such a message can be obtained by a call to
Stunnel_reauthenticate(). The resulting token will be a good authentication
test which unpacks to zero bytes in Stunnel_retrieve().
Stunnel_is_authenticated() tells wether a valid token already was received.
If Stunnel_reauthenticate() is used, one has to be aware that Stunnel_retrieve()
might return 2 without having an unfinished token. Use Stunnel_has_bytes()
to find out wether an unfished token is being processed if this information
is of interest for your application. 

If the receiving peer is a kind of server which only replys on request, then
a special initiating message is not necessary. The first request will do.


Several Keys:

A StunneL has up to four SblowfisH encryption objects using different keys.

The internal key is hardcoded into the program. It is used to read and write
keyfiles. Also it is used to generate the random generator key from current
time and some other data. 
This key is set by the macro Stunnel_internal_keY and should be altered if
it is possible to keep secret the programs that use StunneL. This adds a
great deal of security, if possible.

The random key is used to generate random numbers. Such numbers are needed to
generate keys and challenges. Also they help to obscure sequence numbers
with their high redundancy (which is the enemy of all encryption).

The user's personal key is read from a file where it is stored encrypted by
the programs internal key. It is used for a short while during the handshake
until the session key is transmitted. Therefore it does not wear off quickly.

The session key is generated at random (hopefully) during handshake. As soon
as it is transmitted, all encryption is done with this key.  


*/

#endif /* ! Stunnel_includeD */

